#!/usr/bin/python
# This is a runnable Python file.
# Running it and seeing the output is a key part of what it has to say.

"""If you can understand everything in this file, you understand Python.
If you can write any of the things in this file, you can write Python.

By the way, this long triple-quoted-string at the top of the file,
assigned to no variable, is a 'documentation' string. Python's built-in
help system can recognize it as describing the contents of the file.
Documentation strings can also be put just inside class and function
definitions. More on that later.
"""

# I coment behind single hash characters, for the rest of the line.
# Comments have no effect on the program, and are just for people reading the code.
# These comments can go many places.

a_number = 5  # Here I just created a variable named 'a_number' with the value 5
a_integer = 3
a_float = 2.7

tuple_of_numbers = (1, 5, 7)
list_of_numbers = [2, 8, 32]
list_of_strings = ['a', 'b', 'tomato']
string_in_single_quotes = 'Maybe I want to put a "quotation" inside a string;'
string_in_double_quotes = "or make it easy to put single quotes in the string's body."
long_string = """Either way, the above two strings behave the same way.
It's just a style or convention issue, which do you want to use.
If you write lots of HTML, you might prefer
'single quote strings with <a href="#tag">tags</a> in them'.
This long triple-quoted-string can contain either ' or " or ''' and also contains
all
these
line
brakes
and isn't done until the closing """

print """
# Accessing the elements of tuples and lists.
# Looping.
"""

print tuple_of_numbers[1], 'is 5 because lists and tuples start counting with the zeroeth element.'
print list_of_numbers[1], 'is 8. lists and tuples look the same when accessing one element.'
# This is 'iterating' over the contents of lists and tuples:
print 'tuple_of_numbers:'
for i in tuple_of_numbers:
	print i
print 'list_of_numbers:'
for i in list_of_numbers:
	print i
# What just happened there is that a new variable 'i' was successively assigned
# each of the elements of the tuple or list. One of the above lists is the same as:
print 'tuple_of_numbers:'
i = tuple_of_numbers[0]
print i
i = tuple_of_numbers[1]
print i
i = tuple_of_numbers[2]
print i
# In addition to being smaller to write, the for-loop knows the length of the thing
# it is iterating over and stops appropriately.
# Here it is again as a 'while' loop.
print 'list_of_numbers:'
position = 0
while position < len(list_of_numbers):
	# If position is less than the length of the list_of_numbers, do the following:
	print list_of_numbers[position]
	position = position + 1
	# Here, at the end of the while loop, we go back to the top and consider the
	# 'while condition' again. If true, do the 'loop body', the above two
	# statements, again.
# We checked for less-than the length, because we start counting from 0.
# As spelled out above in the unrolling of printing tuple_of_numbers, if there
# are three elements, the valid element positions are 0, 1 and 2, but not 3.
#
# Let's break down that loop one more time:
print 'list_of_numbers:'
position = 0
while True:
	# Loop forever!
	print list_of_numbers[position]
	# "a += b" is shorthand for "a = a + b"
	position += + 1
	if position >= len(list_of_numbers):
		# the break statement jumps to outside the enclosing loop and moves us past it.
		break


print """
# You can't assign into tuples.
# try-except
"""

list_of_numbers[1] = 9
print 'I can assign to a list element and change it\'s value to:', list_of_numbers
list_of_numbers.append(88)
print 'I can also append to lists using the append member function of the list object (more on that later)', list_of_numbers
print 'print understands how to print many Python objects, separated by commas:', tuple_of_numbers

try:
	# assigning to tuple elements isn't allowed, they're frozen the way they're created
	tuple_of_numbers[1] = 99
except Exception, e:
	# but inside of a try-except statement we can catch when things go wrong,
	# maybe report the error, maybe do something else,
	# maybe give up and exit the program.
	print 'Here follows the Exception, e:'
	print e
	print '---- thus spake the Exception'

print """
# Dictionaries
"""

# Dictionaries are collections of key-value pairs.
# You can set and get values based on some key value.
dict_of_values = {
'key one': 'value one',
'key two': 2,
3: 'value three'
}
print """dict_of_values['key one'] =""", dict_of_values['key one']
print """dict_of_values['key two'] =""", dict_of_values['key two']
print """dict_of_values[3] =""", dict_of_values[3]
try:
	print dict_of_values['key not in dict']
except Exception, e:
	print 'trying to get the value of a dict with no such key!', e
if 'key not in dict' in dict_of_values:
	print 'you can check to see if a key is "in" a dictionary, but'
else:
	print 'it is not there'

dict_of_values['new key'] = 'new value'
print dict_of_values
# Note in the output, the key:value pairs get printed in no particular order.
# Lists and tuples have order, but dictionaries don't.

print """
# Formatting
"""
# print is kinda coarse, using % formatting is much more flexible, and often
# easier to understand for putting together complex things.
print 'percent-s is a string %s, percent-d is an integer %d, percent-f is a float %f' % ('a string', 55, 123.456)
# In general, 'format string' % (tuple, of, values, to, substitute, into, the, format)
print 'many Python objects can be converted to a string "%s" with the repr() function.' % (repr(tuple_of_numbers))


print """
# Functions
"""

def addOne(some_number):
	"""This is a pretty trivial function, it takes one value and returns it plus 1.
	
	(psst, documentation string!).
	"""
	return some_number + 1

print 5, ' -> addOne -> ', addOne(5)

def sum(something_iterable):
	"""You should write your own functions to group together chunks of program
	that make sense to go together, or will be run many times, possibly from
	different parts of your greater program."""
	out = 0
	for x in something_iterable:
		# Note that we don't need to know if something_iterable is a list or
		# a tuple (or something else entirely that Python treats as 'iterable'!)
		out += x
	return out

print 'tuple_of_numbers:', tuple_of_numbers, ' -> sum -> ', sum(tuple_of_numbers)
print 'list_of_numbers:', list_of_numbers, ' -> sum -> ', sum(list_of_numbers)

print """
# import
"""

# import statements usually go at the top of a file,
# but I didn't want to introduce them up there yet.
# An import statement says I need to use something else in a Python module,
# in this case the 'math' module of the Python standard library.
import math

# math has useful constants like pi
print 'Pi is about %f' % (math.pi)
# and functions like sin and sqrt
print 'sin(%f) = %f' % (math.pi/4, math.sin(math.pi/4))
print 'sqrt(2)/2 = %f' % (math.sqrt(2)/2)

# The 'sys' package of the standard library has lots of basic parts of the
# system as Python sees it, including, where does print go?
import sys

print 'print is just like sys.stdout.write, sorta'
sys.stdout.write('sys.stdout.write is just like print, sorta\n')


print """
# Classes, Objects
"""

class IveGotClass(object):
	"""I've mentioned 'python objects' a few times, by which I mean things
	more general than just lists and tuples and dictionaries
	(but those things are also python objects).
	Here I'm defining a new type of object called "IveGotClass" to show
	how some of these things work by default, and how you can make these
	things for your own usage."""
	
	def __init__(self):
		"""The __init__ method is a function that is a special part of
		a class. It is the first thing called when one of these objects
		is being created. It can take more arguments that can be used
		to set initial values. It always takes at least the 'self' argument
		which is an object of type IveGotClass that we're working with
		right now."""
		self.instance_variable = 'initial value'
		# This is the first time I'm using the special 'None' value
		# which means that the variable has no value.
		# None is different than an empty string '' or an empty list []
		# or an empty dict {} or empty tuple () or 0.
		self.another = None
		self.set_count = 0
	
	def getValue(self):
		"""The simplest kind of methods are getters and setters, but often
		much happens inside of methods to maintain an internal state of values
		inside the object instance."""
		return self.instance_variable

	def setValue(self, new_value):
		"""In addition to just setting the value, we count how many times
		it has been set. This is a toy example of what can be done."""
		self.instance_variable = new_value
		self.set_count += 1
	
	def __str__(self):
		"""__str__ is a special method that returns a string that is a nice
		human readable representation of what this object is. You'll see it
		used below, by print and % magic!"""
		return 'IveGotClass("%s", set_count=%d)' % (self.instance_variable, self.set_count)

# here we have two instances of IveGotClass
foo = IveGotClass()  # this calls __init__ to set things up
bar = IveGotClass()

print 'foo', foo.getValue()
print 'bar', bar.getValue()

print 'set value ...'
bar.setValue('new value')

print 'foo', foo.getValue()
print 'bar', bar.getValue()

# How does Python know how to print objcts? Depends on their __str__ method!
print foo
print bar
